%%
%% %CopyrightBegin%
%% 
%% Copyright Ericsson AB 2009-2019. All Rights Reserved.
%% 
%% Licensed under the Apache License, Version 2.0 (the "License");
%% you may not use this file except in compliance with the License.
%% You may obtain a copy of the License at
%%
%%     http://www.apache.org/licenses/LICENSE-2.0
%%
%% Unless required by applicable law or agreed to in writing, software
%% distributed under the License is distributed on an "AS IS" BASIS,
%% WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
%% See the License for the specific language governing permissions and
%% limitations under the License.
%% 
%% %CopyrightEnd%
%%
-module(snmpa_discovery_handler).

-export([verify/1]).

-callback stage1_finish(TargetName, ManagerEngineID, ExtraInfo) ->
    ignore |
    {ok, snmpa_conf:usm_entry() | [snmpa_conf:usm_entry()]} |
    {ok, snmpa_conf:usm_entry() | [snmpa_conf:usm_entry()], NewExtraInfo} when
      TargetName      :: string(),
      ManagerEngineID :: string(),
      ExtraInfo       :: term(),
      NewExtraInfo    :: term().
      
verify(Mod) ->
    snmp_misc:verify_behaviour(?MODULE, Mod).

